from __future__ import print_function
from __future__ import absolute_import
from builtins import object
import logging
from logging.handlers import RotatingFileHandler, WatchedFileHandler
import os
import errno
from csr_cloud.as_aws import as_cloud
from csr_cloud.ha_csr import HACsr

GUESTSHELL = "/home/guestshell/"
LOGS = "logs/"
CLOUD = "cloud/"
LOG = ".log"
FORSLASH = "/"
AWS = "aws"
AUTOSCALER = "autoscaler"
HA = "ha"
EVENTS = "events"


class csr_cloud(object):
    def __init__(self, feature, input1=None, input2=None):
        self.cloud_name = AWS

        self.feature = feature.lower()

        # Create the logger
        self.log = logging.getLogger(self.feature)

        if self.feature == AUTOSCALER:
            self.as_cloud = as_cloud()

        elif self.feature == HA:
            self.setup_logging()
            # Make event-log directory
            event_directory = GUESTSHELL + CLOUD  + FORSLASH + self.feature.upper() + FORSLASH + EVENTS
            self.mkdir_p(event_directory)
            self.ha = HACsr()

    def mkdir_p(self, path):
        try:
            os.makedirs(path, exist_ok=True)  # Python>3.2
        except TypeError:
            try:
                os.makedirs(path)
            except OSError as exc:
                if exc.errno == errno.EEXIST and os.path.isdir(path):
                    pass
                else:
                    raise

    def setup_logging(self):
        try:
            # Name the directory
            directory_name = GUESTSHELL + CLOUD  + FORSLASH + self.feature.upper()
            path = directory_name + FORSLASH
            # Create the directory
            self.mkdir_p(path)

            # Name the log file
            log_file = 'csr_ha.log'

            if self.feature == HA:
                handler = WatchedFileHandler(filename=os.path.join(directory_name, log_file), mode='a')

            else:

                # Create the file handler
                handler = RotatingFileHandler(os.path.join(directory_name, log_file), mode='a', maxBytes=5 * 1024 * 1024,
                                              backupCount=0, encoding=None, delay=0)

            # Create the format
            formatter = logging.Formatter(
                ' %(asctime)s  %(module)s:  %(funcName)s:%(lineno)s  %(levelname)s  %(message)s')

            # Set the format
            handler.setFormatter(formatter)

            # Add Handler
            if not len(self.log.handlers):
                self.log.addHandler(handler)

            # Set the log level
            self.log.setLevel(logging.DEBUG)

        except Exception as e:
            print("csr_cloud: setup_logging: exception {}. ".format(e))
            pass


if __name__ == '__main__':
    csr = csr_cloud('HA')


import boto3
import datetime
import logging
import ast
import os
from botocore.client import Config
from collections import namedtuple
import re
import string
import json
import time
from xml.dom import minidom

logger = logging.getLogger('autoscaler')

#This function adds a <transit_vpc_config /> block to an existing XML doc and returns the new XML
def updateConfigXML(xml, config, vgwTags, account_id, csr_number): # pragma: no cover
    xmldoc=minidom.parseString(xml)
    #Create TransitVPC config xml block
    transitConfig= xmldoc.createElement("transit_vpc_config")
    #Create Account ID xml block
    newXml = xmldoc.createElement("account_id")
    newXml.appendChild(xmldoc.createTextNode(account_id))
    transitConfig.appendChild(newXml)

    #Create VPN Endpoint xml block
    newXml = xmldoc.createElement("vpn_endpoint")
    newXml.appendChild(xmldoc.createTextNode(csr_number))
    transitConfig.appendChild(newXml)

    #Create status xml block (create = tagged to create spoke, delete = tagged as spoke, but not with the correct spoke tag value)
    newXml = xmldoc.createElement("status")
    newXml.appendChild(xmldoc.createTextNode("create"))
    transitConfig.appendChild(newXml)

    # Configure preferred transit VPC path
    # TO-DO: Where to get preferred_path from now if I want to support that, a param in CFN template?
    # Maybe I just repurpose vgwTags arg and instead pass PreferredPathTag all the way down?
    newXml = xmldoc.createElement("preferred_path")
    if vgwTags is not None:
        newXml.appendChild(xmldoc.createTextNode(vgwTags.get(config.get('PREFERRED_PATH_TAG', 'none'), 'none')))
    else:
        newXml.appendChild(xmldoc.createTextNode('none'))
    transitConfig.appendChild(newXml)

    #Add transit config to XML
    xmldoc.childNodes[0].appendChild(transitConfig)
    return str(xmldoc.toxml())

def createRandomPassword(pwdLength=15, specialChars="False"):
    logger.info("Creating random password")
    if specialChars is None:
        specialChars = "True"
    # Generate new random password
    chars = string.ascii_letters + string.digits
    if specialChars == "True":
        chars += '#$%^&+='
        p = re.compile(
            '^(?=.{1,})(?=.*[0-9])(?=.*[a-z])(?=.*[A-Z])(?=.*[#$%^&+=]).*$')
    else:
        p = re.compile('^(?=.{1,})(?=.*[0-9])(?=.*[a-z])(?=.*[A-Z]).*$')
    numTries = 0
    pwdFound = False
    while not pwdFound:
        password = ''
        numTries += 1
        for i in range(int(pwdLength)):
            password += chars[ord(os.urandom(1)) % len(chars)]
        m = p.match(password)
        if m is not None:
            pwdFound = True
    logger.info("Password created after %s tries", numTries)
    logger.debug("%s", password)
    return password


def create_userdata(cfg, status, inst_num, fingerprint, autoscaler_tag):
    bgp_asn = status.get_bgp_asn()
    hostname = cfg.get_as_group_name() + '-' + str(inst_num)
    username = "automate"
    password = createRandomPassword()
    userdata = []
    userdata.append("license=\"%s\"\n" % cfg.get_license_level_type())

    userdata.append("ios-config-1=\"username %s priv 15 pass %s\n" %
                    (username, password))
    userdata.append("ios-config-2=\"service password-encryption\"\n")
    userdata.append("ios-config-3=\"crypto isakmp policy 200\"\n")
    userdata.append("ios-config-4=\"encryption aes 128\"\n")
    userdata.append("ios-config-5=\"authentication pre-share\"\n")
    userdata.append("ios-config-6=\"group 2\"\n")
    userdata.append("ios-config-7=\"lifetime 28800\"\n")
    userdata.append("ios-config-8=\"hash sha\"\n")
    userdata.append(
        "ios-config-9=\"crypto ipsec transform-set ipsec-prop-vpn-aws esp-aes 128 esp-sha-hmac\"\n")
    userdata.append("ios-config-10=\"mode tunnel\"\n")
    userdata.append("ios-config-11=\"crypto ipsec df-bit clear\"\n")

    userdata.append(
        "ios-config-12=\"crypto isakmp keepalive 10 10 on-demand\"\n")

    userdata.append(
        "ios-config-13=\"crypto ipsec security-association replay window-size 1024\"\n")
    userdata.append(
        "ios-config-14=\"crypto ipsec fragmentation before-encryption\"\n")
    userdata.append("ios-config-15=\"crypto ipsec profile ipsec-vpn-aws\"\n")
    userdata.append("ios-config-16=\"set pfs group2\"\n")
    userdata.append(
        "ios-config-17=\"set security-association lifetime seconds 3600\"\n")
    userdata.append("ios-config-18=\"set transform-set ipsec-prop-vpn-aws\"\n")

    userdata.append("ios-config-23=\"ip ssh pubkey-chain\"\n")
    userdata.append("ios-config-24=\"username %s %s" % (username, "\"\n"))

    userdata.append("ios-config-25=\"key-hash ssh-rsa %s %s" %
                    (fingerprint, "\"\n"))
    userdata.append(
        "ios-config-26=\"ip ssh server algorithm authentication publickey\"\n")
    userdata.append("ios-config-27=\"ip ssh maxstartups 1\"\n")

    userdata.append("ios-config-170=\"hostname %s\"\n" % hostname)
    userdata.append("ios-config-171=\"line vty 0 20\"\n")
    userdata.append("ios-config-172=\"login local\"\n")
    userdata.append("ios-config-173=\"transport input ssh\"\n")
    userdata.append("autoscaler-tag=\"%s\"\n" % autoscaler_tag)


    # userdata.append("ios-config-180=\"call-home\"\n")
    # userdata.append("ios-config-181=\"profile CiscoTAC-1\"\n")
    # userdata.append("ios-config-182=\"active\"\n")
    # userdata.append("ios-config-183=\"destination transport-method http\"\n")
    # userdata.append("ios-config-184=\"no destination transport-method email\"\n")

    s = ''.join(userdata)

    return s


class as_cloud():
    def __init__(self):
        self.logger = logging.getLogger('autoscaler')

        self.cloudwatch = boto3.client('cloudwatch')
        self.region = os.getenv('AWS_DEFAULT_REGION', "us-east-1")

        # todo: check to see that region is probably setup for lambda
        self.endpoint_url = {
            "us-east-1": "https://s3.amazonaws.com",
            "us-east-2": "https://s3-us-east-2.amazonaws.com",
            "us-west-1": "https://s3-us-west-1.amazonaws.com",
            "us-west-2": "https://s3-us-west-2.amazonaws.com",
            "eu-west-1": "https://s3-eu-west-1.amazonaws.com",
            "eu-central-1": "https://s3-eu-central-1.amazonaws.com",
            "ap-northeast-1": "https://s3-ap-northeast-1.amazonaws.com",
            "ap-northeast-2": "https://s3-ap-northeast-2.amazonaws.com",
            "ap-south-1": "https://s3-ap-south-1.amazonaws.com",
            "ap-southeast-1": "https://s3-ap-southeast-1.amazonaws.com",
            "ap-southeast-2": "https://s3-ap-southeast-2.amazonaws.com",
            "sa-east-1": "https://s3-sa-east-1.amazonaws.com"
        }

        self.s3_resource = boto3.resource('s3', endpoint_url=self.endpoint_url[self.region], config=Config(
            s3={'addressing_style': 'virtual'}, signature_version='s3v4'))
        self.s3_client = boto3.client('s3', endpoint_url=self.endpoint_url[self.region], config=Config(
            s3={'addressing_style': 'virtual'}, signature_version='s3v4'))
        self.sns = boto3.client('sns')
        self.ec2 = boto3.resource('ec2')
        self.ec2_client = boto3.client('ec2')
        self.events_client = boto3.client('events')

        self.Results = namedtuple('Results', 'min, max, avg, cnt, inst')

        self.privatekey_prefix = "AutoScaler/privatekeys/"


    def get_storage_prefix(self):
        if 'bucket_prefix' in os.environ:
            vpnconfig_prefix = os.environ['bucket_prefix']
        else:
            vpnconfig_prefix = "vpnconfigs/"

        return vpnconfig_prefix

    def get_as_private_key_files_info(self, instance_id):
        '''
        This method returns transit network solutions' private key files location information
        :param instance_id (str):
        :return: key_tuple (tuple object) - ( private key directory path, file name)
        '''
        return (self.privatekey_prefix, str(instance_id) + '.pem')

    def get_short_instance_id(self, instance_id):
        '''
        :param instance_id:
        :return: instance_id
        '''
        return instance_id

    def get_tnet_private_key_files_info(self, instance_id):
        '''
        This method returns Autoscaler solutions' private key files location information
        :param instance_id (str):
        :return: key_tuple (tuple object) - ( private key directory path, file name)
        '''
        return (self.get_storage_prefix(), 'prikey.pem')

    def get_transit_network_filename(self):
        return "transit_vpc_config.txt"

    def get_transit_network_config(self, storage_name, config_filename):
        vpnconfig_prefix = self.get_storage_prefix()
        self.logger.info("storage_name: %s" % storage_name)
        self.logger.info("vpnconfig_prefix: %s" % vpnconfig_prefix)
        self.logger.info("config_filename: %s" % config_filename)
        for i in range(5):
            try:
                config = ast.literal_eval(
                    self.s3_client.get_object(
                        Bucket=storage_name,
                        Key=vpnconfig_prefix +
                        config_filename)['Body'].read())
            except Exception as e:
                self.logger.warning("Exception while getting vpc config: %s" % e)
                config = None
                continue
        return config


    def put_transit_network_Config(self, status, storage_name, config_file, config):
        vpnconfig_prefix = self.get_storage_prefix()

        self.logger.info(
            "Uploading new config file: %s/%s/%s%s",
            self.endpoint_url[self.region],
            storage_name,
            vpnconfig_prefix,
            config_file)
        try:
            self.s3_client.put_object(
                Bucket=storage_name,
                Key=vpnconfig_prefix +
                config_file,
                Body=str(config),
                ACL='bucket-owner-full-control',
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=status.get_kms_key())

        except Exception as e:
            self.logger.warning("Failed to put VPC config file: %s" % e)
            return False

        return True

    def get_instanceId_from_ip(self, public_ip_address):
        try:
            addresses_dict = self.ec2_client.describe_addresses(
                PublicIps=[public_ip_address])
            return addresses_dict["Addresses"][0]['InstanceId']
        except Exception as e:
            self.logger.warning("Exception while looking up ip address: %s" % e)
            return None

    def get_instance_config_filename(self, instance_id):
        return "%s.cfg" % instance_id

    def get_private_key(self, storage_name, prikey):
        if os.path.exists('/tmp/' + prikey):
            os.remove('/tmp/' + prikey)

            self.logger.debug(
            "Downloading private key: %s/%s/%s%s",
            self.endpoint_url[self.region],
            storage_name,
            self.privatekey_prefix,
            prikey)

        try:
            self.s3_client.download_file(
                storage_name, self.privatekey_prefix + prikey, '/tmp/' + prikey)
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        return True


    def put_private_key(self, status, storage_name, instance_id, prikey):

        instance_private_key_filename = instance_id + '.pem'
        self.logger.debug(
            "Downloading private key: %s/%s/%s%s",
            self.endpoint_url[self.region],
            storage_name,
            self.privatekey_prefix,
            prikey)
        try:
            self.s3_client.put_object(
                Bucket=storage_name,
                Key=self.privatekey_prefix +
                instance_private_key_filename,
                Body=str(prikey),
                ACL='bucket-owner-full-control',
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=status.get_kms_key())
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False
        return True


    def copy_private_key(self, status, storage_name, from_file, to_file):

        try:
            self.s3_resource.Object(
                storage_name,
                self.privatekey_prefix +
                to_file).copy_from(
                CopySource=storage_name +
                '/' +
                self.privatekey_prefix +
                from_file,
                ACL='bucket-owner-full-control',
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=status.get_kms_key())
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        return True

    def copy_original_privateKey(self, status, storage_name, from_file, to_file):
        vpnconfig_prefix = self.get_storage_prefix()

        try:
            self.s3_resource.Object(
                storage_name,
                self.privatekey_prefix +
                to_file).copy_from(
                CopySource=storage_name +
                '/' +
                vpnconfig_prefix +
                'prikey.pem',
                ACL='bucket-owner-full-control',
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=status.get_kms_key())
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        return True


    def save_original_privateKey(self, status, storage_name, from_file, to_file):
        vpnconfig_prefix = self.get_storage_prefix()

        try:
            self.s3_resource.Object(
                storage_name,
                vpnconfig_prefix +
                'prikey.pem').copy_from(
                CopySource=storage_name +
                '/' +
                self.privatekey_prefix +
                from_file,
                ACL='bucket-owner-full-control',
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=status.get_kms_key())
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        return True


    def associate_role_to_instance(self, instance_id, cloud_config):
        role = cloud_config["iam_instance_profile"]
        try:
            self.ec2_client.associate_iam_instance_profile(
                IamInstanceProfile={
                    'Arn': role,
                },
                InstanceId=instance_id
            )
        except Exception as e:
            self.logger.error(
                "Not able to associate %s to %s -> %s" %
                (role, instance_id, e))
            return False

        return True

    def set_instance_tags(self, cfg, instance_id, name):
        try:
            i = self.ec2.Instance(instance_id)

            self.ec2.create_tags(Resources=[i.id], Tags=[
                {'Key': 'Name', 'Value': name}])

            self.ec2.create_tags(Resources=[i.id], Tags=[
                {'Key': 'AutoScaleGroup', 'Value': cfg.get_as_group_name()}])

            self.ec2.create_tags(Resources=[i.id], Tags=[
                {'Key': 'storage_name', 'Value': cfg.get_storage_name()}])

        except Exception as e:
            self.logger.warning(
                "Failed to set label on instance %s to %s (%s)" %
                (instance_id, name, e))
            return False

        return True

    def is_instance_ready(self, instance_id):
        try:
            response = self.ec2_client.describe_instance_status(
                Filters=[
                    {
                        'Name': 'instance-status.reachability',
                        'Values': [
                            'passed',
                        ]
                    },
                ],
                InstanceIds=[
                    instance_id,
                ],
            )
            if len(response['InstanceStatuses']):
                instance_status = response['InstanceStatuses'][0]['InstanceStatus']['Status']
            else:
                instance_status = "Not ready."
        except Exception as e:
            instance_status = "Not Ready : %s" % e

        self.logger.info("Instance %s is %s" % (instance_id, instance_status))

        if instance_status == "ok":
            return True
        else:
            return False


    def is_instance_running(self, instance_id):
        try:
            response = self.ec2_client.describe_instances(
                InstanceIds=[
                    instance_id,
                ],
            )
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        if len(response['Reservations']) == 0:
            return False

        for r in response['Reservations']:
            for i in r['Instances']:
                if i['State']['Name'] == 'running':
                    return True
        return False


    def get_running_state(self, instance_id):
        try:
            response = self.ec2_client.describe_instances(
                InstanceIds=[
                    instance_id,
                ],
            )
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return "terminated"

        if len(response['Reservations']) == 0:
            return "terminated"

        # 'Name': 'pending' | 'running' | 'shutting-down' | 'terminated' | 'stopping' | 'stopped'
        for r in response['Reservations']:
            for i in r['Instances']:
                return i['State']['Name']
        return "terminated"


    def get_reachability_status(self, instance_id):
        try:
            response = self.ec2_client.describe_instance_status(
                InstanceIds=[
                    instance_id,
                ],
            )
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return "Failed"

        if "InstanceStatuses" in response:
            for instance_response in response["InstanceStatuses"]:
                system_status = instance_response['SystemStatus']['Details'][0]['Status']
                instance_status = instance_response['InstanceStatus']['Details'][0]['Status']
                if system_status == "passed" and instance_status == "passed":
                    return "Passed"

        return "Failed"


    def put_file(self, status, storage_name, directory, filename):
        try:
            self.s3_client.upload_file(
                '/tmp/' + filename,
                storage_name,
                directory + filename,
                ExtraArgs={
                    "ServerSideEncryption": "aws:kms",
                    "SSEKMSKeyId": status.get_kms_key()})

        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        return True

    def get_file(self, storage_name, directory, filename):
        try:
            self.s3_client.download_file(
                storage_name,
                directory + filename,
                '/tmp/' + filename)
        except Exception as e:
            self.logger.warning(
                "Did not find %s:%s%s (%s)" %
                (storage_name, directory, filename, e))
            return False

        return True


    def does_file_exist(self, storage_name, filename):
        if filename is None:
            return False
        try:
            if self.s3_client.list_objects(
                    Bucket=storage_name,
                    Prefix='AutoScaler' + filename):
                return True

            return False

        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False


    def copy_file(self, status, storage_name, directory, from_file, to_file):

        try:
            self.s3_resource.Object(
                storage_name,
                directory +
                to_file).copy_from(
                CopySource=storage_name +
                '/' +
                directory +
                from_file,
                ACL='bucket-owner-full-control',
                ServerSideEncryption='aws:kms',
                SSEKMSKeyId=status.get_kms_key())

        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        return True

    # stub function to keep the parity between azure utils and aws utils
    def copy_file_from_to_dir(self, storage_name, from_dir, from_file, to_dir, to_file):
        return True

    def get_metric(
            self,
            name,
            instance_id,
            period,
            metric_period,
            metric_type,
            namespace):

        dimensions = [
            {'Name': "InstanceId",
             'Value': instance_id}
        ]
        if metric_type == "Sum":
            stats_type = ["Sum"]
        else:
            stats_type = ["Maximum"]

        response = self.cloudwatch.get_metric_statistics(
            Namespace=namespace,
            MetricName=name,
            # Unit='Count',
            Dimensions=dimensions,
            StartTime=datetime.datetime.utcnow() -
            datetime.timedelta(
                seconds=int(period)),
            EndTime=datetime.datetime.utcnow(),
            Period=metric_period,
            Statistics=stats_type,
        )
        if response['ResponseMetadata']['HTTPStatusCode'] == 200:
            mylist = []
            for dp in response['Datapoints']:
                if metric_type == "Sum":
                    value = (dp[stats_type[0]]) / metric_period
                else:
                    value = dp[stats_type[0]]
                mylist.append((name, dp['Timestamp'], value))

            if len(mylist):
                results = sorted(mylist, key=lambda x: x[0])
                name, min_time, min_value = min(
                    results, key=lambda item: item[2])
                name, max_time, max_value = max(
                    results, key=lambda item: item[2])
                avg_value = (sum(x[2] for x in results) / len(mylist))
                self.logger.debug("%s (%s) => %s/%s/%s" %
                             (instance_id, name, min_value, max_value, avg_value))

                result = self.Results(
                    min=min_value,
                    max=max_value,
                    avg=avg_value,
                    cnt=len(mylist),
                    inst=instance_id)

                return result
            else:
                self.logger.debug("No Datapoints for %s (%s)" % (instance_id, name))
        else:
            self.logger.warning("Bad response: %s" % response)
        return None


    def put_metric(self, name, description, group_name):
        try:
            response = self.cloudwatch.put_metric_data(
                Namespace="csr1000v",
                MetricData=[{'MetricName': name,
                             'Value': 1,
                             'Unit': 'Count',
                             'Dimensions': [
                                 {
                                     'Name': "AutoScaleGroup",
                                     'Value': group_name
                                 },
                                 {
                                     'Name': "Action",
                                     'Value': description
                                 }
                             ]
                             }
                            ],
            )
            if response['ResponseMetadata']['HTTPStatusCode'] != 200:
                self.logger.warning("Bad response for %s: %s" % (name, response))
                return False
            return True

        except Exception as e:
            self.logger.warning("Exception %s" % e)
            return False

        return True

    def send_notification(
            self,
            topicarn,
            msg="",
            subject="",
            msgtype='string',
            message_attributes={}):

        self.logger.info("topicarn: %s" % topicarn)
        self.logger.info("msg: %s" % msg)
        self.logger.info("subject: %s" % subject)
        self.logger.info("message_attributes: %s" % message_attributes)
        try:
            response = self.sns.publish(
                TopicArn=topicarn,
                Subject=subject,
                Message=msg,
                MessageStructure=msgtype,
                MessageAttributes=message_attributes

            )
            self.logger.info("response from sending notification: %s" % response)
        except Exception as e:
            self.logger.warning("Failed to send notification: %s" % e)
            response = e

        self.logger.info("Done sending response")
        return response

    def send_user_notification(self, cloud_config,
                                        msg="",
                                        subject="",
                                        msgtype='string',
                                        message_attributes={}):
        if 'TopicArn' in cloud_config:
            topicarn = cloud_config['TopicArn']
        else:
            self.logger.info("No Notification ARN found in cloud config")
            return "Not Sent"

        self.send_notification(topicarn, msg=msg,
                                        subject=subject,
                                        msgtype='string',
                                        message_attributes=message_attributes)

    def put_rule(self, arn, minutes):

        try:
            response = self.events_client.list_rule_names_by_target(
                TargetArn=arn,
            )
            for rule_name in response['RuleNames']:
                name = rule_name
                break
            else:
                return False
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        if minutes > 1:
            frequency = "rate(%d minutes)" % int(minutes)
        else:
            frequency = "rate(%d minute)" % int(minutes)

            self.logger.info("frequency: %s" % frequency)
        try:
            rule_response = self.events_client.put_rule(
                Name=name,
                ScheduleExpression=frequency,
                State='ENABLED',
            )
            self.logger.info("%s" % rule_response)
        except Exception as e:
            self.logger.warning("Exception putting rule: %s" % e)


    def get_rule_minutes(self, arn):
        try:
            response = self.events_client.list_rule_names_by_target(
                TargetArn=arn,
            )
            for rule_name in response['RuleNames']:
                name = rule_name
                break
            else:
                return None
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False

        try:
            rule_response = self.events_client.describe_rule(
                Name=name
            )
            self.logger.info("%s" % rule_response)
        except Exception as e:
            self.logger.warning("Exception getting rule: %s" % e)
            return 999

        expression = rule_response['ScheduleExpression']
        numbers = re.findall(r'\d+', expression)
        return int(numbers[0])


    def get_cloud_cfg_subset(self, instance_id, default_cfg=None):
        cloud_config = {}
        az_list = []
        subnet_list = []
        sec_group_list = []
        try:
            ec2 = boto3.resource('ec2')

            i = ec2.Instance(instance_id)
            az_list.append(i.placement["AvailabilityZone"])

            for network_interface in i.network_interfaces_attribute:
                subnet_list.append(network_interface["SubnetId"])
                for group in network_interface['Groups']:
                    sec_group_list.append(group["GroupId"])

            cloud_config["instance_type"] = i.instance_type
            # todo: get appropriate image
            cloud_config["ami-id"] = i.image_id
            cloud_config["key_name"] = i.key_name
            cloud_config["az_list"] = az_list
            cloud_config["subnet_id"] = subnet_list
            cloud_config["security_groups"] = sec_group_list
            if i.iam_instance_profile is not None:
                cloud_config["iam_instance_profile"] = i.iam_instance_profile['Arn']
            else:
                cloud_config["iam_instance_profile"] = "Unknown"
                self.logger.warning(
                    "Currently, instance %s does not have an instance profile" %
                    instance_id)

            cloud_config["public_ip_address"] = i.public_ip_address
            cloud_config["private_ip_address"] = i.private_ip_address
            self.logger.debug(
                "Instance %s config =\n %s\n" %
                (instance_id, cloud_config))
            return cloud_config
        except Exception as e:
            self.logger.warning("Failed to get config for %s (%s)" % (instance_id, e))
            return default_cfg


    def get_cloud_config(self, instance_id):
        return self.get_cloud_cfg_subset(instance_id)

    def get_public_ip(self, instance_id, saved_config):
        if saved_config and "public_ip_address" in saved_config:
            return saved_config["public_ip_address"]
        else:
            config = self.get_cloud_config(instance_id)
            return config["public_ip_address"]

    def get_image_id(self, instance_id, saved_config):
        if saved_config and "ami-id" in saved_config:
            return saved_config["ami-id"]
        else:
            config = self.get_cloud_config(instance_id)
            return config["ami-id"]

    def get_ip_allocation_id(self, instance_id):
        config = self.get_cloud_config(instance_id)
        ip_address = config["public_ip_address"]
        try:
            response = self.ec2_client.describe_addresses(
                PublicIps=[
                    ip_address
                ],
            )
            return response["Addresses"][0]['AllocationId']
        except:
            return None


    def reserve_ip(self, name=None):
        try:
            eip = self.ec2_client.allocate_address(Domain='vpc')
            elastic_ip = eip['PublicIp']
            allocationId = eip["AllocationId"]
        except Exception as e:
            self.logger.info("Elastic IP was not created, Error: %s ", e)
            elastic_ip = None
            allocationId = None

        return elastic_ip, allocationId


    def associate_ip(self, instance_id, allocationId):
        try:
            self.ec2_client.associate_address(
                InstanceId=instance_id,
                AllocationId=allocationId)
        except Exception as e:
            self.logger.info("Could not associate ip for %s (%s)" % (instance_id, e))
            return False

        return True

    def disassociate_ip(self, instance_id, allocation_id):
        try:
            self.ec2_client.disassociate_address(
                AssociationId=allocation_id
            )

        except Exception as e:
            self.logger.info("Could not disassociate ip for %s (%s)" % (instance_id, e))
            return False

        return True

    def release_ip(self, instance_id):
        allocationid = self.get_ip_allocation_id(instance_id)
        if allocationid is None:
            self.logger.info("Can not release ip for instance %s" % instance_id)
            return False

        self.disassociate_ip(instance_id, allocationid)
        try:
            eip = self.ec2_client.release_address(
                AllocationId=allocationid)
        except Exception as e:
            self.logger.exception("Exception: %s" % e)
            return False
        self.logger.info("Eip:%s" % eip)
        return True


    def spin_up_csr(self, cfg, status, image_id, name, allocationId, fingerprint, tag):
        inst_num = int(name.rsplit('-', 1)[1])
        self.logger.info("Instance Number: %d" % inst_num)

        ud = create_userdata(cfg, status, inst_num, fingerprint, tag)
        logger.info("Userdata: %s", ud)

        cloud_config = cfg.get_default_instance_config()
        self.logger.info("Cloud Config: %s" % cloud_config)

        num_az = len(cloud_config["az_list"])
        which_az = (inst_num-1) % num_az
        az_zone = cloud_config["az_list"][which_az]
        subnetId = cloud_config["subnet_id"][which_az]
        self.logger.info(
            "num_az %d, which_az %d, az_zone %s" %
            (num_az, which_az, az_zone))


        iam_instance_profile = {}
        iam_instance_profile['Arn'] = cloud_config["iam_instance_profile"]

        self.logger.info("Cloud Config: %s" % cloud_config)



        try:
            csr_list = self.ec2.create_instances(
                ImageId=image_id,
                InstanceType=cloud_config["instance_type"],
                KeyName=cloud_config["key_name"],
                MaxCount=1,
                MinCount=1,
                Placement={
                    'AvailabilityZone': az_zone
                },
                SecurityGroupIds=[cloud_config["security_groups"]],
                SubnetId=subnetId,
                UserData=ud,
                DisableApiTermination=False,
                IamInstanceProfile=iam_instance_profile,
                InstanceInitiatedShutdownBehavior='stop',
            )
        except Exception as e:
            self.logger.exception("EXCEPTION: %s" % e)
            return None, None

        csr = csr_list[0]

        csr.wait_until_running()

        self.set_instance_tags(cfg, csr.id, name)

        response = self.ec2_client.modify_instance_attribute(
            SourceDestCheck={
                'Value': False
            },
            InstanceId=csr.id
        )

        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            self.logger.warning("Bad response for SourceDestCheck: %s" % response)

        if cloud_config['DisableApiTermination'] == "Yes":
            response = self.ec2_client.modify_instance_attribute(
                DisableApiTermination={
                    'Value': True
                },
                InstanceId=csr.id
            )

            if response['ResponseMetadata']['HTTPStatusCode'] != 200:
                self.logger.warning(
                    "Bad response for DisableApiTermination: %s" %
                    response)

        # todo: what if the following fails
        csr.wait_until_running()

        return csr.id


    def spin_down_csr(self, instance_list, config):
        waiter = self.ec2_client.get_waiter('instance_terminated')

        for instance_id in instance_list:
            try:
                instance = self.ec2.Instance(instance_id)
                # if termination protection is enabled, disable it before
                # terminating.
                response = instance.describe_attribute(
                    Attribute='disableApiTermination')
                if response['DisableApiTermination']['Value']:
                    response = instance.modify_attribute(DisableApiTermination={
                        'Value': False
                    })
                    self.logger.info("%s" % response)

                response = instance.terminate()
                self.logger.info("%s" % response)

            except Exception as e:
                self.logger.error("Failed to Terminate: %s" % e)
                return False

        waiter.wait(InstanceIds=instance_list)

        return True


    def reboot_instance(self, instance_id):
        try:
            response = self.ec2_client.reboot_instances(
                InstanceIds=[
                    instance_id
                ],
            )

            self.logger.info("Rebooted instance %s" % instance_id)
            self.logger.debug("%s" % response)

        except Exception as e:
            self.logger.error("Failed to reboot instance %s: %s" % (instance_id, e))
            return False

        return True


    def clean_storage(self, storage_name):
        bucket = self.s3_resource.Bucket(storage_name)

        for s3_file in bucket.objects.filter(Prefix=self.privatekey_prefix):
            instance_id = (os.path.basename(s3_file.key).split('.')[0])
            if self.get_running_state(instance_id) == "terminated":
                try:
                    self.s3_client.delete_object(
                        Bucket=storage_name,
                        Key=s3_file.key,
                    )
                    self.logger.info("Deleted %s file" % (s3_file.key))
                except Exception as e:
                    self.logger.exception("Exception: %s" % e)
                    pass
            else:
                self.logger.info(
                    "Instance %s is not terminated.  No need to delete pem file" %
                    instance_id)


    def update_function(self, function_name, storage_name, key_name):
        client = boto3.client('lambda')

        try:
            response = client.update_function_code(
                FunctionName=function_name,
                S3Bucket=storage_name,
                S3Key=key_name,
                # S3ObjectVersion='string',
                Publish=True,
                # RevisionId='string'
            )
        except Exception as e:
            self.logger.error("Failed to upgrade function %s: %s" % (function_name, e))
            return False

        self.logger.info("Updated function %s" % response)
        return True

    def is_csr_image_version_16_9_1(self, image_id):
        try:
            response = self.ec2_client.describe_images(
                Filters=[
                    {
                        'Name': 'image-id',
                        'Values': [
                            image_id,
                        ]
                    },
                ]
            )
            image_name = response['Images'][0]['Name']
            self.logger.info("Image name is %s" % image_name)
            if "cisco-CSR-." in image_name:
                parts = image_name.split("cisco-CSR-.")
                version = parts[1].split('.')
                major = version[0]
                minor = version[1]
                point = version[2].split('-')[0]
                if major == '16' and minor == '09' and point == '01a':
                    self.logger.info("Version 16.09.01a image. Returning True")
                    return True
                else:
                    self.logger.info("Version is different than 16.09.01a. Returning False")
                    return False
            else:
                self.logger.info("Private image or image name does not contain 'cisco-CSR' string. Returning False")
                return False

        except Exception as e:
            self.logger.exception("Version not found in image name. Exception is %s \nReturning False" % e)
            return False

    def configure_dmvpn_hub(self, status, instance_id):
        '''
        Get dmvpn_hub_config.txt from S3 bucket and write to a jon object
        Add CSR + instance number Eg. CSR1 to the json object
        Get Lambda ARN from dmvpn_hub_config.txt
        invoke lambda to configure CSR as DMVPN hub
        '''
        if self.is_spoke_cloud_vpn_gateway:
            logger.info("No need to configure CSR as DMVPN Hub as spoke vpn device is cloud gateway\n")
            return
        try:
            lambda_client = boto3.client('lambda')
            bucket_name = status.storage_name
            s3_prefix = self.get_storage_prefix()
            dmvpnConfig = ast.literal_eval(
                self.s3_client.get_object(Bucket=bucket_name, Key=s3_prefix + 'dmvpn_hub_config.txt')[
                    'Body'].read())
            lambda_function = dmvpnConfig['dmvpn_lambda']
            csr_num = status.get_instance_number(instance_id)
            payload = {}
            payload['dmvpnParams'] = dmvpnConfig
            payload['csr'] = 'CSR%d' % csr_num
            tmp = {}
            eip = status.get_instance_ip(instance_id)
            pip = status.get_instance_private_ip(instance_id)
            tmp['EIP%d' % csr_num] = eip
            tmp['PIP%d' % csr_num] = pip
            self.s3_client.put_object(
            Bucket=bucket_name,
            Key=s3_prefix + 'scale_out_csr_config_tmp_{}.txt'.format(csr_num),
            Body=str(tmp),
            ACL='bucket-owner-full-control',
            ServerSideEncryption='aws:kms',
            SSEKMSKeyId=status.get_kms_key())
            payload['eip'] = eip
            logger.info("Payload: {}".format(payload))
            res = lambda_client.invoke(FunctionName=lambda_function,
                                       InvocationType='RequestResponse',
                                       Payload=json.dumps(payload))
            logger.info("Configure Hub Lambda response: {}".format(res))
        except Exception as e:
            logger.exception("Exception in configure_dmvpn_hub: {}".format(e))

    def is_spoke_cloud_vpn_gateway(self, status):
        '''
        :param status: autoscaler controller's snapshot of current state machine
        :return: True if spoke VPN gateway is TGW else returns False
        '''
        config_file = "transit_vpc_config.txt"
        config = self.get_transit_network_config(status.storage_name, config_file)
        if 'TgwId' in config:
            logger.info("Spoke is a TGW.\n")
            return True
        else:
            logger.info("Spoke is a CSR.\n")
            return False

    def create_vpn_config(self, status, instance_id):
        '''
        Create VPN connection with spoke TGW so that scaled out CSR comes up with Ipsec tunnels with TGW
        :param status: autoscaler controller's snapshot of current state machine
        :param instance_id: scaled_out instance id
        :return:
        '''
        try:
            bucket_name = status.storage_name
            bucket_prefix = self.get_storage_prefix()
            config_file = self.get_transit_network_filename()
        
            # Retrieve Transit VPC configuration from transit_vpn_config.txt
            logger.info('Getting config file %s/%s%s', bucket_name, bucket_prefix, config_file)
            config = self.get_transit_network_config(bucket_name, config_file)

            tgw_id = config['TgwId']
            account_id = config['AccountId']
            tgw_vpn_attachment_id = ""
            tgw_route_tbl_id = config['TgwRouteTableId']
            vpn_state = 'pending'
            csr_num = status.get_instance_number(instance_id)

            # The workflow is as follows:
            # - Create Customer Gateway
            # - Create VPN connection (which might take some time to go from pending to available state)
            # - Push VPN connection config to S3 bucket (which triggers Configurator lambda)
            # - Get available VPN attachment
            # - Associate TGW attachment and enable propagation for TGW route table
            # Since lambda has a timeout of 5 minutes, if any of the above operations take a long time to complete,
            # the lambda function might time out, leading to an in-consistent state.
            # Saving state after each of the above steps to overcome lambda timeout issue
            if 'CSR%d_vpn_id' % csr_num not in config:
                logger.info('Creating VPN attachments for TGW %s', tgw_id)
                # Create Customer Gateways (will create CGWs if they do not exist, otherwise, the API calls are ignored)
                logger.debug('Creating Customer Gateways with IP %s' % config['EIP%d' % csr_num])
                cg1 = self.ec2_client.create_customer_gateway(Type='ipsec.1', PublicIp=config['EIP%d' % csr_num],
                                                              BgpAsn=config['BGP_ASN'])
                self.ec2_client.create_tags(Resources=[cg1['CustomerGateway']['CustomerGatewayId']],
                                            Tags=[{'Key': 'Name', 'Value': 'Transit VPC Endpoint1'}])

                logger.info('Created Customer Gateways: %s' % cg1['CustomerGateway']['CustomerGatewayId'])
                time.sleep(10)
                # Create and tag first VPN connection
                vpn1 = self.ec2_client.create_vpn_connection(Type='ipsec.1',
                                                             CustomerGatewayId=cg1['CustomerGateway'][
                                                                 'CustomerGatewayId'],
                                                             TransitGatewayId=tgw_id,
                                                             Options={'StaticRoutesOnly': False})
                self.ec2_client.create_tags(Resources=[vpn1['VpnConnection']['VpnConnectionId']],
                                            Tags=[
                                                {'Key': 'Name', 'Value': tgw_id + '-to-Transit-VPC CSR%s' % csr_num},
                                                {'Key': config['HUB_TAG'], 'Value': config['HUB_TAG_VALUE']},
                                                {'Key': 'transitvpc:endpoint', 'Value': 'CSR%d' % csr_num}
                                            ])

                vpn_conn_id = vpn1['VpnConnection']['VpnConnectionId']
                logger.info('Created VPN connections: %s' % vpn_conn_id)
                logger.info('Writing VPN Information in transit_vpc_config.txt file and uploading to S3.\n')
                config['CSR%d_vpn_id' % csr_num] = vpn_conn_id
                self.put_transit_network_Config(status, bucket_name, self.get_transit_network_filename(), config)

            else:
                vpn_conn_id = config['CSR%d_vpn_id' % csr_num]
        
            while (vpn_state is not 'available'):
                # Retrieve VPN configuration
                vpn_config1 = self.ec2_client.describe_vpn_connections(
                    VpnConnectionIds=[vpn_conn_id])
                vpn_state = vpn_config1['VpnConnections'][0]['State']
                if vpn_state == 'available':
                    logger.info("VPN {} state is available\n".format(vpn_conn_id))
                    break
                logger.info("VPN {} state is {}. Retrying after 30 seconds\n".format(vpn_conn_id, vpn_state))
                time.sleep(30)

            vpn_config1 = vpn_config1['VpnConnections'][0]['CustomerGatewayConfiguration']

            # Retrieve Transit VPC configuration from transit_vpn_config.txt
            logger.info('Getting config file %s/%s%s', bucket_name, bucket_prefix, config_file)
            config = self.get_transit_network_config(bucket_name, config_file)

            if 'CSR{}_{}_config_file'.format(csr_num, vpn_conn_id) not in config:
                # Update VPN configuration XML with transit VPC specific configuration info for this connection
                vpn_config1 = updateConfigXML(vpn_config1, config, None, account_id, 'CSR%d' % csr_num)
                
                # Put CSR1 config in S3
                self.s3_client.put_object(
                    Body=str.encode(vpn_config1),
                    Bucket=bucket_name,
                    Key=bucket_prefix + 'CSR%d/' % csr_num + self.region + '-' + vpn_conn_id + '.conf',
                    ACL='bucket-owner-full-control',
                    ServerSideEncryption='aws:kms',
                    SSEKMSKeyId=config['KMS_KEY']
                )
                
                logger.info('Pushed VPN configurations to S3...')
                config['CSR{}_{}_config_file'.format(csr_num, vpn_conn_id)] = 1
                self.put_transit_network_Config(status, bucket_name, self.get_transit_network_filename(), config)

            num_vpns = 0
            while (num_vpns < 1):
                tgw_vpn_attachments = self.ec2_client.describe_transit_gateway_attachments(
                    Filters=[{'Name': 'transit-gateway-id', 'Values': [tgw_id]},
                             {'Name': 'resource-type', 'Values': ['vpn']}, {'Name': 'state', 'Values': ['available']}])

                for tgw_attach in tgw_vpn_attachments['TransitGatewayAttachments']:
                    tgw_vpn_attachment_id = tgw_attach['TransitGatewayAttachmentId']
                    logger.info('Found TGW VPN attachment with ID %s!', tgw_vpn_attachment_id)
                    if "Association" in tgw_attach:
                        logger.info(
                            "Attachment {} is already associated with TGW route table\n".format(tgw_vpn_attachment_id))
                        tgw_vpn_attachment_id = ""
                    else:
                        num_vpns = len(tgw_vpn_attachments['TransitGatewayAttachments'])
                        break

                if num_vpns >= 1:
                    break
                logger.info("Retrying after 30 seconds\n")
                time.sleep(30)

            if tgw_vpn_attachment_id:
                # Create a TGW association and propagation for each TGW VPN attachment
                logger.info('Associating TGW attachment with ID %s to TGW route table!', tgw_vpn_attachment_id)
                tgw_assoc = self.ec2_client.associate_transit_gateway_route_table(
                    TransitGatewayRouteTableId=tgw_route_tbl_id,
                    TransitGatewayAttachmentId=tgw_vpn_attachment_id)
                logger.info('Enabling propagation for TGW attachment with ID %s to TGW route table!',
                            tgw_vpn_attachment_id)
                tgw_prop = self.ec2_client.enable_transit_gateway_route_table_propagation(
                    TransitGatewayRouteTableId=tgw_route_tbl_id,
                    TransitGatewayAttachmentId=tgw_vpn_attachment_id)

                logger.info('Created TGW association: {}'.format(tgw_assoc))
                logger.info('Created TGW association and route propagation for CSR%d VPN attachments!' % csr_num)
        except Exception as e:
            logger.exception("Exception in create_vpn_config: {}".format(e))
            raise



import setuptools
from distutils.core import setup

project_name = 'csr_aws_utils'
project_ver = '2.0.0'

setup(
    name=project_name,
    packages=["csr_cloud"],  # this must be the same as the name above
    install_requires=[
        'boto3==1.14.9', 
        'requests==2.23.0'
    ],
    version=project_ver,
    description='Utilities for csr1000v on AWS',
    author='Christopher Reder',
    author_email='creder@cisco.com',
    # use the URL to the github repo
    url='https://github4-chn.cisco.com/csr1000v-aws/' + project_name,
    download_url='https://github4-chn.cisco.com/csr1000v-aws/' + project_name + '/archive/' + \
        project_ver + '.tar.gz',
    keywords=['cisco', 'aws', 'guestshell', 'csr1000v'],
    classifiers=[],
    license="MIT"
)



'''
    This file will be used for functionality defining to the HA feature in aws
'''

from builtins import str
from builtins import object
import logging
from boto3 import client, resource
import datetime
import os
import sys
import argparse
import re
import json
import requests
import time
import shutil

DESTINATION_CIDR_BLOCK = "DestinationCidrBlock"
NETWORK_INTERFACE_ID = "NetworkInterfaceId"
ROUTE_TABLE_ID = "routeTableId"
REGION = "region"
ENI_ID = "eniId"
ROUTE = "route"
MODE = "mode"
PRIMARY = "primary"
HA = 'ha'
EVENTS = 'events'
EVENT_LOGS = '/home/guestshell/cloud/HA/events'
REVERT = 'revert'

class HACsr(object):
    '''
     Functions implemented as api's from csr_ha package
    '''

    def __init__(self):
        self.log = logging.getLogger('ha')
        site=subprocess.check_output("{} -m site --user-site".format(sys.executable), shell=True)
        sys.path.append(site + b'/csr_ha/client_api')
        sys.path.append(site + b'/csr_ha/server')

        # Setting event_logger for route events for fail-safe approach
        # Ideally csr_ha is responsible for setting it up using set_event_logger
        self.event_logger = self.log
        self.event_log_file = ""
        try:
            if os.path.exists("/home/guestshell/cloud/HA/csr_ha.log") and os.path.getsize("/home/guestshell/cloud/HA/csr_ha.log") > 15000000:
                if os.path.exists("/bootflash/guest-share"):
                    shutil.copy('/home/guestshell/cloud/HA/csr_ha.log', '/bootflash/guest-share/csr_ha.log')
                else:
                    shutil.copy("/home/guestshell/cloud/HA/csr_ha.log", "/bootflash/")
                os.remove('/home/guestshell/cloud/HA/csr_ha.log')
        except Exception as e:
            #This is backing up the csr_ha.log file, if file size is above 10 MB. Log the exception and continue so that HACsr object initialization doesn't fail because of this
            self.log.exception("Exception in init is {e}".format(e=e))
            self.log.info("Backing up csr_ha.log to cloud/HA/csr_ha_backup.log")
            shutil.move("/home/guestshell/cloud/HA/csr_ha.log", "/home/guestshell/cloud/HA/csr_ha_backup.log")

        res = ''
        num_retry = 0
        while res == '':
            try:
                num_retry += 1
                res = requests.get("http://169.254.169.254/latest/dynamic/instance-identity/document")
                self.log.info("CSR is in region: %s", res.json()['region'])
                self.region = res.json()['region']
            except Exception as e:
                if num_retry == 5:
                    self.log.warning('Error in fetching region after 5 retries. Exception is %s' % e)
                    self.region = 'us-east-1'
                    self.log.warning("Region is set to us-east-1 by default")
                    pass
                else:
                    self.log.warning('Error in fetching region...Retrying')
                    self.region = 'us-east-1'
                    time.sleep(10)
                    continue

        ec2_client = client('ec2', region_name=self.region)

        self.log.info("HACsr initialization complete")

    def validate_route_table_string(self, s, pat=re.compile(r"rtb-[a-fA-F0-9]{8,}")):
        self.log.info("Validating if the route table id format matches with AWS")
        if not pat.match(s) or len(pat.match(s).group(0)) != len(s):
            raise argparse.ArgumentTypeError('input %s does not match the route table id pattern "rtb-[a-fA-F0-9]{8,}"' % s)
        return s

    def validate_eni_string(self, s, pat=re.compile(r"eni-[a-fA-F0-9]{8,}")):
        self.log.info("Validating if the ENI ID format matches with aws")
        if not pat.match(s) or len(pat.match(s).group(0)) != len(s):
            raise argparse.ArgumentTypeError('input %s does not match the eni id pattern "eni-[a-fA-F0-9]{8,}"' % s)
        return s

    def validate_inputs(self, args):
        node = {}
        node[REGION] = args.rg
        response, fail_response_code = self.get_response_dict()
        resp_msg = ""

        if args.rg:
            node[REGION] = args.rg
            self.log.info("Validating Region in AWS cloud. This validation is being done using the region {r}".format(r=self.region))
            try:
                ec2_client = client('ec2', region_name=self.region)
                regions = ec2_client.describe_regions(
                    RegionNames=[
                        node[REGION]
                    ]
                )
                self.log.info("Region found in AWS: %s" % regions)
            except Exception as e:
                self.log.warning("\nWarning: Error in validating region. Exception is {e}".format(e=e))
                resp_msg = resp_msg + "\nWarning: Error in validating region. Exception is {e}".format(e=e)
                response['code'] = fail_response_code

        if args.t or args.r:
            node[ROUTE_TABLE_ID] = args.t
            self.log.info("Reading route table {r} in region {g}".format(r=args.t, g=args.rg))

            rt = self.get_route_table_using_ec2_client(node)
            if rt is None:
                self.log.warning("\nWarning: Route table {r} not found in region {g}".format(r=args.t, g=args.rg))
                resp_msg = resp_msg + "\nWarning: Route table {r} not found in region {g}\n".format(r=args.t, g=args.rg)
                response['code'] = fail_response_code
                if args.r:
                    self.log.warning("\nWarning: Route cannot be validated since Route table does not exist")
                    resp_msg = resp_msg + "\nWarning: Route cannot be validated since Route table does not exist"
            else:
                if args.r:
                    node[ROUTE] = args.r
                    route_found = 0
                    routes = rt['RouteTables'][0]['Routes']
                    for route in routes:
                        if 'DestinationCidrBlock' in route:
                            if route['DestinationCidrBlock'] == node[ROUTE]:
                                route_found = 1
                        if 'DestinationIpv6CidrBlock' in route:
                            if route['DestinationIpv6CidrBlock'] == node[ROUTE]:
                                route_found = 1

                    if route_found == 0:
                        self.log.warning("\nWarning: Route {r} not found in Route Table {t}".format(r=args.r, t=args.t))
                        resp_msg = resp_msg + "\nWarning: Route {r} not found in Route Table {t}".format(r=args.r, t=args.t)
                        response['code'] = fail_response_code
        if args.n:
            node[ENI_ID] = args.n
            eni = self.get_eni_id(node)
            if not eni:
                self.log.warning("\nWarning: ENI ID {e} not found in region {g}".format(e=args.n, g=args.rg))
                resp_msg = resp_msg + "\nWarning: ENI ID {e} not found in region {g}".format(e=args.n, g=args.rg)
                response['code'] = fail_response_code

        if response['code'] == 200:
            resp_msg = resp_msg + "\nAll inputs are valid\n"
        else:
            resp_msg = resp_msg + "\nPlease verify the node configuration"
        response['msg'] = resp_msg
        return json.dumps(response)

    def validate_set_param_inputs(self, node, new_params):
        self.log.info("Validating if set node parameters are valid")
        resp_msg = ""
        response, fail_response_code = self.get_response_dict()
        # for key in new_params:
        self.log.info(new_params)
        if ROUTE_TABLE_ID in list(new_params.keys()) or ROUTE in list(new_params.keys()):
            self.log.info("Validating if Route Table exists")
            rt = self.get_route_table_using_ec2_client(node)
            if not rt:
                if ROUTE_TABLE_ID in list(new_params.keys()):
                    self.log.warning("\nWarning: Route table {r} not found. Please check node configuration again.".format(r=new_params[ROUTE_TABLE_ID]))
                    resp_msg = resp_msg + "\nWarning: Route table {r} not found. Please check node configuration again.".format(r=new_params[ROUTE_TABLE_ID])
                    response['code'] = fail_response_code
                if ROUTE in list(new_params.keys()):
                    self.log.warning("\nWarning: Route cannot be validated since Route table does not exist")
                    resp_msg = resp_msg + "\nWarning: Route cannot be validated since Route table does not exist"
                    response['code'] = fail_response_code
            else:
                if ROUTE in list(new_params.keys()):
                    self.log.info("Validating if route is valid")
                    route_found = 0
                    routes = rt['RouteTables'][0]['Routes']
                    for route in routes:
                        if 'DestinationCidrBlock' in route:
                            if route['DestinationCidrBlock'] == node[ROUTE]:
                                route_found = 1
                        if 'DestinationIpv6CidrBlock' in route:
                            if route['DestinationIpv6CidrBlock'] == node[ROUTE]:
                                route_found = 1

                    if route_found == 0:
                        self.log.warning("\nWarning: Route {r} not found. Please check node configuration again.".format(r=new_params[ROUTE]))
                        resp_msg = resp_msg + "\nWarning: Route {r} not found. Please check node configuration again.".format(r=new_params[ROUTE])
                        response['code'] = fail_response_code

        if ENI_ID in list(new_params.keys()):
            self.log.info("Validating if ENI ID is valid")
            eni = self.get_eni_id(node)
            if not eni:
                self.log.warning("\nWarning: ENI ID {e} not found. Please check node configuration again.".format(e=new_params[ENI_ID]))
                resp_msg = resp_msg + "\nWarning: ENI ID {e} not found. Please check node configuration again.".format(e=new_params[ENI_ID])
                response['code'] = fail_response_code

        if REGION in list(new_params.keys()):
            self.log.info("Validating Region in AWS cloud. This validation is being done using the region {r}".format(r=self.region))
            ec2_client = client('ec2', region_name=self.region)
            try:
                regions = ec2_client.describe_regions(
                    RegionNames=[
                        node[REGION]
                    ]
                )
            except Exception as e:
                self.log.warning("\nWarning: Region {g} not found. Please check node configuration again.".format(g=new_params[REGION]))
                resp_msg = resp_msg + "\nWarning: Region {g} not found. Please check node configuration again.".format(g=new_params[REGION])
                response['code'] = fail_response_code

        if response['code'] == 200:
            resp_msg = resp_msg + "All inputs are valid"
        else:
            resp_msg = resp_msg + "\nPlease verify the node configuration"

        response['msg'] = resp_msg
        return response

    def create_node(self, params):
        '''
            This node will get the additional params from the metadata server if they are not provided
        '''
        self.log.debug("creating node {p}".format(p=params))

        # At this point we don't need to modify any params but its possible that we add it in the future
        return params

    def create_node_parser(self, optional, required):
        '''
            This function will create a parser for the create_node python file in the csr_ha package
            Also includes the aws specific parameters needed
        '''

        # Required arguments

        required.add_argument('-t', help='<routetable-id> rtb-xxxxxxxx', default=None, required=True, type=self.validate_route_table_string)
        required.add_argument('-rg', help='<region> eg: us-east-1', default=None, required=True)
        required.add_argument('-n', help='<eni-id of peer> eni-xxxxxxxx', default=None, required=True, type=self.validate_eni_string)

        # Optional arguments
        optional.add_argument('-r', help='<destination cidr block> x.x.x.x/x', default=None, required=False)

    def set_node_parser(self, optional, required):
        '''
            This function will create a parser for set_node python file in the csr_ha package
        '''

        # Optional arguments
        self.create_optional_arguments(optional)

    def clear_param_parser(self, optional, required):
        '''
            This function will return a parser for clear_params python file in the csr_ha package
        '''
        # Optional arguments
        curr_action = "store_true"

        # Optional arguments
        optional.add_argument('-t', help='<routetable-id> rtb-xxxxxxxx', default=None, required=False,
                              action=curr_action)
        optional.add_argument('-rg', help='<region> eg: us-east-1', default=None, required=False,
                              action=curr_action)
        optional.add_argument('-n', help='<eni-id> eni-xxxxxxxx', default=None, required=False,
                              action=curr_action)
        optional.add_argument('-r', help='<destination cidr block> x.x.x.x/x', default=None, required=False,
                              action=curr_action)

    def check_clear_param(self, index, old_params):
        '''
            This function will return the keyword for the parameter to be deleted
        '''
        self.log.debug("check_clear_param index: {i}, params: {p}".format(i=index, p=old_params))

        keyword = self.check_param_command(old_params[index])

        return keyword

    def verify_node(self, node, event_type):
        '''
            This function will verify if the required parameters are set for a particular node for that event
        '''
        param_str = " "
        self.event_logger.info("Redundancy node configuration: ")
        for key in node:
            param_str = param_str + "{:15s}{} \n".format(key, node[key])

        self.event_logger.info(param_str)

        event_type = event_type.lower()

        # verify if for that node all parameters have been setup correctly
        if event_type == "verify" or event_type == "peerfail" or event_type == "revert":

            node_verified = True
            if ROUTE_TABLE_ID not in node:
                node_verified = False
                self.event_logger.warning("route table id has not been configured")
            if REGION not in node:
                node_verified = False
                self.event_logger.warning("region has not been configured")
            if ENI_ID not in node:
                node_verified = False
                self.event_logger.warning("eni id has not been configured")
            if ROUTE not in node:
                self.event_logger.warning("Destination cidr block has not been configured, will set all-routes")

            if node_verified is True:
                self.event_logger.info("All required parameters have been configured")
                return 'OK'
            else:
                return 'ERR1'
        else:
            self.event_logger.error("Unknown event type: {e}".format(e=event_type))
            return 'ERR1'

    def get_response_dict(self, event=None):
        # if the request is coming from validate_inputs, the return codes are successful response (200) and warning
        # response (199)
        # if the request is coming from any event for validation, the return codes are successful response (200) and
        # failed response (400) because we want to take an action based on these return codes
        response = {}
        resp_msg = ""
        success_return_code = 200
        response['code'] = success_return_code
        fail_return_code = 199
        if event:
            fail_return_code = 400
        return response, fail_return_code

    def get_eni_id(self, node, event_type=None):
        '''
        This function returns a boto3 eni object
        '''
        eni = None
        try:
            ec2_client = client('ec2', region_name=node[REGION])
            eni = ec2_client.describe_network_interfaces(
                NetworkInterfaceIds=[
                    node[ENI_ID],
                ]
            )
        except Exception as e:
            self.event_logger.exception("Exception: {ex}".format(ex=e))

        return eni

    def get_route_table_using_ec2_client(self, node):
        '''
        ec2 resource object returns the route table object even if it does not exist, which is why for validation we need to use client object
        '''
        route_table = None
        try:
            ec2_client = client('ec2', region_name=node[REGION])
            route_table = ec2_client.describe_route_tables(
                RouteTableIds=[
                    node[ROUTE_TABLE_ID],
                ]
            )
        except Exception as e:
            self.event_logger.warning("Exception: {ex}".format(ex=e))

        return route_table

    def get_route_table(self, node, event_type=None):
        '''
            This function returns a route-table boto3 resource object
        '''
        route_table = None

        try:
            if event_type and event_type == 'verify':
                self.event_logger.info("Fetching the route table")
            ec2_resource = self.get_ec2_resource(node[REGION])
            route_table = ec2_resource.RouteTable(node[ROUTE_TABLE_ID])
        except Exception as e:
            self.event_logger.exception("Exception: {ex}".format(ex=e))

        if route_table is None:
            self.event_logger.info("Route table not found: {}".format(ROUTE_TABLE_ID))
            if event_type and event_type == 'verify':
                self.event_logger.info("It is likely permission to access the route table was not granted.")
        else:
            if event_type == 'verify':
                param_str = ""
                self.event_logger.info("VpcId: {r}".format(r=route_table.vpc_id))
                self.event_logger.info("RouteTableId: {r}".format(r=route_table.route_table_id))
                self.event_logger.info("Routes in the route table: ")

                route_list = route_table.routes_attribute
                for route in route_list:
                    if 'DestinationCidrBlock' in route:
                        self.event_logger.info("CIDR: {r}".format(r=route['DestinationCidrBlock']))
                        cidr_block = route['DestinationCidrBlock']
                    if 'DestinationIpv6CidrBlock' in route:
                        self.event_logger.info("v6CIDR: {r}".format(r=route['DestinationIpv6CidrBlock']))
                        cidr_block = route['DestinationIpv6CidrBlock']
                    if 'NetworkInterfaceId' in route:
                        self.event_logger.info("ENI: {r}".format(r=route['NetworkInterfaceId']))
                        next_hop = route['NetworkInterfaceId']
                    elif 'GatewayId' in route:
                        self.event_logger.info("GW: {r}".format(r=route['GatewayId']))
                        next_hop = route['GatewayId']

                    param_str += "{:15s}{} \n".format(cidr_block, next_hop)

                self.event_logger.info(param_str)

        return route_table

    def set_route_table(self, node, event_type, route_table):
        '''
            This function will update the route table for different event types
        '''
        self.event_logger.debug("Setting route table for node: {n} , event_type: {e}".format(
                n=node, e=event_type))
        response_code = 400
        try:

            if route_table is None:
                self.event_logger.error("Route table object is not set, cannot update the table")
                return None

            send_request = False

            # Event type code exists for csr_ha to control the behavior of set_route_table function in the future
            event_type = event_type.lower()
            if event_type == "verify":
                self.event_logger.debug("route_table: {r}".format(r=route_table))
                send_request = True
            elif event_type == "peerfail":
                send_request = True
            elif event_type == "revert":
                if MODE in node:
                    if node[MODE] == PRIMARY:
                        send_request = True
            else:
                self.event_logger.error("Incorrect event type: {e}".format(e=event_type))

                return None

            responses = []

            if send_request is True:
                # Get all the routes from the route table
                routes = route_table.routes_attribute
                next_hop = None
                if routes:
                    if ROUTE in node:
                        # Set one specific route
                        response_code = self.set_one_route(node, route_table, event_type)
                    else:
                        # Set all routes
                        response_code = self.set_all_routes(node, routes, event_type)
            else:
                self.delete_log_file_revert(event_type)

            return response_code

        except Exception as e:
            self.event_logger.exception("Exception: {ex}".format(ex=e))

    def check_cloud_command(self, cmd):
        '''
            This function checks the cloud commands passed through to the cloud specific layer
        '''

        keyword = self.check_param_command(cmd)
        return keyword

    def validate_nexthop(self, nexthop):
        '''
            This function will return True if csr_ha has to validate nexthop ip otherwise return False
        '''

        # For aws, currently next hop is going to be only an ENI-ID, thus we don't have an ip to validate
        return False

    def set_event_logger(self, node, event_type, directory_name=EVENT_LOGS):
        '''
            This function will set the event_logger for ha
        '''

        self.event_logger = self.create_event_logger(node, event_type, directory_name)

    '''
        Functions implemented as helpers to the above api's
    '''

    def get_ec2_resource(self, region):
        '''
            This function returns the boto3 ec2 resource
        '''

        ec2_resource = resource('ec2', region_name=region)
        return ec2_resource

    def create_optional_arguments(self, optional, store_values=False):
        '''
            This function will return the optional arguments for the parser
        '''

        curr_action = None

        if store_values:
            curr_action = "store_true"

        # Optional arguments
        optional.add_argument('-t', help='<routetable-id> rtb-xxxxxxxx', default=None, required=False,
                              action=curr_action, type=self.validate_route_table_string)
        optional.add_argument('-rg', help='<region> eg: us-east-1', default=None, required=False,
                              action=curr_action)
        optional.add_argument('-n', help='<eni-id> eni-xxxxxxxx', default=None, required=False,
                              action=curr_action, type=self.validate_eni_string)
        optional.add_argument('-r', help='<destination cidr block> x.x.x.x/x', default=None, required=False,
                              action=curr_action)

    def check_param_command(self, cmd):
        '''
            This function is a common function for check_cloud_command and check_clear_param
        '''

        keyword = ""
        if cmd == '-t':
            keyword = ROUTE_TABLE_ID
        elif cmd == '-rg':
            keyword = REGION
        elif cmd == '-n':
            keyword = ENI_ID
        elif cmd == '-r':
            keyword = ROUTE
        else:
            self.log.error("Invalid parameter/command format {c}".format(c=cmd))
            return "Error"

        return keyword

    def set_one_route(self, node, route_table, event_type):
        '''
            This function will set the given route to a given next hop eni
        '''

        # Update the route for the given route cidr only if the eni_id is not our own
        response = []
        sent_request = False
        rsp_code = 400

        routes = route_table.routes_attribute
        for route in routes:
            self.event_logger.info("In set_one_route Route {}".format(route))
            ipv6_route =  False
            if 'DestinationIpv6CidrBlock' in route:
                cidr_block = route['DestinationIpv6CidrBlock']
                ipv6_route = True
            elif 'DestinationCidrBlock' in route:
                cidr_block = route['DestinationCidrBlock']

            if cidr_block == node[ROUTE]:
                if 'NetworkInterfaceId' in route:
                    current_next_hop = route['NetworkInterfaceId']
                    ec2_client = client('ec2', region_name=node[REGION])
                    if current_next_hop != node[ENI_ID] and event_type != 'verify':
                        self.event_logger.info("Replacing the network interface id {} for route: {}".format(node[ENI_ID], cidr_block))
                        if ipv6_route:
                            resp = ec2_client.replace_route(RouteTableId=node[ROUTE_TABLE_ID], DestinationIpv6CidrBlock=cidr_block, NetworkInterfaceId=node[ENI_ID])
                        else:
                            resp = ec2_client.replace_route(RouteTableId= node[ROUTE_TABLE_ID], DestinationCidrBlock=cidr_block, NetworkInterfaceId=node[ENI_ID])
                        response.append(resp)
                        sent_request = True
                    elif event_type == 'verify':
                        self.event_logger.info("Verifying the network interface id {} for route {}".format(current_next_hop, cidr_block))
                        if ipv6_route:
                            resp = ec2_client.replace_route(RouteTableId=node[ROUTE_TABLE_ID], DestinationIpv6CidrBlock=cidr_block, NetworkInterfaceId=current_next_hop)
                        else:
                            resp = ec2_client.replace_route(RouteTableId= node[ROUTE_TABLE_ID], DestinationCidrBlock=cidr_block, NetworkInterfaceId=current_next_hop)
                        response.append(resp)
                        sent_request = True
                        self.event_logger.info("Response for verify event: {}".format(response))
                    elif event_type == 'peerFail':
                        self.event_logger.info("Route is already updated. No write performed.")
                break

        # Delete log file if it was just a revert event where we did not send a request out
        if not sent_request:
            self.delete_log_file_revert(event_type)

        if event_type == 'verify':
            self.event_logger.info("Set route(s) response")
            self.event_logger.info(response)

        # The response is a list of 1 entry. The entry is a dictionary with 1 key.
        # The value for the 1 key is a dictionary with 4 keys.
        if response:
            rsp_dict = response[0]
            rsp_value = rsp_dict['ResponseMetadata']
            rsp_code = rsp_value['HTTPStatusCode']
            if rsp_code == 200:
                if event_type == 'verify':
                    self.event_logger.info("Verify event completed successfully")
                else:
                    self.event_logger.info("Set route completed successfully")
            else:
                if event_type == 'verify':
                    self.event_logger.info("Verify route failed with error %d" % rsp_code)
                else:
                    self.event_logger.info("Set route failed with error %d" % rsp_code)
                    self.event_logger.info(response)

        return rsp_code

    def set_all_routes(self, node, routes, event_type):
        '''
            In case we don't have a CIDR, then we need to set the entire route table to the given next hop eni
        '''
        return_code = 200
        if event_type == "verify":
            sent_request = True
        else:
            sent_request = False

        # Get a boto3 client
        ec2_client = client('ec2', region_name=node[REGION])

        # Try to replace all routes irrespective of any errors that you might get.
        # Note: AWS does not allow you to modify the local route to point to an eni-id
        for route in routes:
            response = []
            try:
                # Only update the routes which don't have a gateway id (local or igw)
                if 'GatewayId' in route:
                    self.event_logger.warning("Cannot replace route for gatewayid: {r}".format(r=route['GatewayId']))
                elif 'NetworkInterfaceId' in route:
                    # Check if the network ID in the node is different from the ID in the route
                    if route['NetworkInterfaceId'] != node[ENI_ID]:
                        if event_type == "verify":
                            # Don't change the next hop for the route
                            next_hop = route['NetworkInterfaceId']
                        else:
                            # Replace the next hop interface with the value in the node
                            next_hop = node[ENI_ID]

                        if 'DestinationCidrBlock' in route:
                            self.event_logger.info("Updating {} with nextHop {} for the {} event".format(route['DestinationCidrBlock'], next_hop, event_type))
                            response = ec2_client.replace_route(RouteTableId=node[ROUTE_TABLE_ID], DestinationCidrBlock=route['DestinationCidrBlock'], NetworkInterfaceId=next_hop)
                        elif 'DestinationIpv6CidrBlock' in route:
                            self.event_logger.info("Updating {} with nextHop {} for the {} event".format(route['DestinationIpv6CidrBlock'], next_hop, event_type))
                            response = ec2_client.replace_route(RouteTableId=node[ROUTE_TABLE_ID], DestinationIpv6CidrBlock=route['DestinationIpv6CidrBlock'], NetworkInterfaceId=next_hop)
                        else:
                            self.event_logger.info("Destination CIDR block not found in route")

                        sent_request = True
                        if event_type == "verify":
                            self.event_logger.info("Set route response")
                            self.event_logger.info(response)
                    else:
                        # Network interface in the route is the same as in the node
                        # No need to update route table entry
                        if 'DestinationCidrBlock' in route:
                            self.event_logger.info("Nexthop to {} already set to {}, no update required".format(route['DestinationCidrBlock'], route['NetworkInterfaceId']))
                        elif 'DestinationIpv6CidrBlock' in route:
                            self.event_logger.info("Nexthop to {} already set to {}, no update required".format(route['DestinationIpv6CidrBlock'], route['NetworkInterfaceId']))
                else:
                    self.event_logger.error("Route without gateway or network interface ignored")

                # The response is a dictionary with 1 key.
                # The value for the 1 key is a dictionary with 4 keys.
                if response:
                    rsp_value = response['ResponseMetadata']
                    rsp_code = rsp_value['HTTPStatusCode']
                    if rsp_code == 200:
                        if event_type == 'verify':
                            self.event_logger.info("Route verify completed successfully")
                        else:
                            self.event_logger.info("Set route completed successfully")
                    else:
                        response_code = rsp_code
                        if event_type == 'verify':
                            self.event_logger.info("Verify route failed with error %d" % rsp_code)
                        else:
                            self.event_logger.info("Set route failed with error %d" % rsp_code)
                            self.event_logger.info(response)

            except Exception as e:
                self.event_logger.exception("Could not update the route in the route table: {exp}".format(exp=e))

        # Delete log file if it was just a revert event where we did not send a request out
        if not sent_request:
            self.delete_log_file_revert(event_type)

        return return_code

    def get_cidr_block(self, route):
        '''
            This function will return the cidr block depending on whether the given address is ipv4 or ipv6
        '''

        if route is None:
            return ""

        cidr_block = ""

        if 'DestinationCidrBlock' in route:
            cidr_block = route['DestinationCidrBlock']
        elif 'DestinationIpv6CidrBlock' in route:
            cidr_block = route['DestinationIpv6CidrBlock']

        return cidr_block


    def create_event_logger(self, node, event_type, directory_name):
        '''
            This function will create the event logger
        '''

        # Create the logger for events
        logger = logging.getLogger("ha.events")

        # Name the log file
        log_file = "node_" + str(node['index']) + "_" + str(
            datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")) + "_" + event_type

        # save the file path so that we can remove the file in-case of unnecessary reverts
        self.event_log_file = os.path.join(directory_name, log_file)

        # Roll over if the number of event files exceeds 513 count
        if len(os.listdir(directory_name)) > 512:
            filetoremove = os.popen("ls -t " + directory_name + "|  tail -1").read().strip()
            self.log.info('Event files threshold exceeded. Removing event file: "{}"'.format(
                os.path.join(directory_name, filetoremove).strip()))
            command = "rm {}".format(os.path.join(directory_name, filetoremove))
            os.system(command)

        # Create the file handler
        handler = logging.FileHandler(filename=os.path.join(directory_name, log_file), mode='a')
        # Add Handler
        if not len(logger.handlers):
            logger.addHandler(handler)

        return logger

    def delete_log_file_revert(self, event_type):
        '''
            This function will delete the file if its a revert and we have elected to not send the
            request out to AWS
        '''

        if event_type == REVERT:
            if os.path.exists(self.event_log_file):
                self.event_logger.info("deleting log file for event_type: {e}, file path: {f}".format(e=event_type, f=self.event_log_file))
                os.remove(self.event_log_file)

    # Taking cloud's input if a revert event is needed if none of nodes are configured as Primary
    def do_revert(self):
        return False

